#include <X11/X.h>
#include <X11/Intrinsic.h>
#include <X11/Xatom.h>
#include <X11/StringDefs.h>
#include <X11/Xutil.h>
#include <X11/Xlib.h>
#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/AtomMgr.h>
#include <Xm/Protocols.h>

extern Display *cx_display;
extern Screen  *cx_screen;
extern Window   cx_rootwindow;

extern Widget   cx_toplevel;
extern Window   cx_topwindow;

/* Use XGetWindowProperty on XA_WM_COMMAND to obtain the Command.
   Remember we get a pointer we must dereference and later release.

   We convert to a single command line by changing nulls to spaces.

   NOTE:
   We should provide client machine,
   at least whenever it is not the server.
   Cannot make it part of the same string because already sized,
   but we can reallocate to make room for it.

   We pass back an array of them for all valid windows.
   Some will be of interest and some may not be.
   The array is static but the the command lines are allocated.
   The caller should free the lines when done. */

#define CX_MAX_COMMAND_LINES 100

static char  *cx_command_lines[CX_MAX_COMMAND_LINES];
static int    cx_cl_count = 0;

static Display *cx_cl_display;

/* This is the recursive call used only internally.
   The user call appears below it. */

static void 
cx_cl_getcmd(Window tree_top, int depth)
{
  Atom      actual_type;
  int       actual_format;
  unsigned long   nitems;
  unsigned long   bytes_leftover;
  unsigned char  *data = NULL;

  Window      root_window, parent_window;
  Window     *child_wins;
  unsigned int  num_children;

  int       chx;
  int       ii;

/* Get past the WM re-parenting but then stop. */

  if (depth > 2)
    return;

/* Get current list of children of tree_top.
   This call returns 0 on failure / 1 on success. */

  if (!XQueryTree(cx_cl_display, tree_top,
          &root_window,
          &parent_window,
          &child_wins,
          &num_children))
    return;

/* If a given window has command property report it.
   Otherwise use recursion to see if a descendant does.
   Direct children of root are still above toplevel! */

  for (chx = 0; chx < num_children; chx++) {

/* Note that XGWP returns 1 on failure / Success on success.
   Our call returns False / True instead.

   Remember that data is one nitems-long character array,
   with internal nulls following each command argument.
   This is NOT the usual argv and argc.
   We convert to a simple command line string. */

    if ((XGetWindowProperty(cx_cl_display, child_wins[chx],
                XA_WM_COMMAND,
                0, 1000, False, AnyPropertyType,
                &actual_type,
                &actual_format,
                &nitems,
                &bytes_leftover,
                &data) == Success)
      && (actual_format == 8)) {
      for (ii = 0; ii < nitems - 1; ii++)
        if (data[ii] == NULL)
          data[ii] = ' ';
      cx_command_lines[cx_cl_count++] = (char *) data;
      cx_command_lines[cx_cl_count] = NULL;
    }

    else {
      if (data)
        XFree(data);

      cx_cl_getcmd(child_wins[chx], depth + 1);
    }
  }
}

/* Call with NULL or empty "" to use existing local display connection;
   or with "host:serve_num" or "host:serve_num.screen_num" for a remote.
   If we open a remote connection we close it when done with it.
   See the XOpenDisplay manpage for all the name variations. */

char      **
cx_getcmd(char *display_name)

{ static char   kws[] = "@(#)cx_getcmd.c	1.1 ARGOS 94/02/05 17:33:55";

  Window      tree_top;

/* Remote if we can.
   Note that DefaultScreen is the one we ask for in the Open. */

  if (display_name
    && strlen(display_name)) {
    if (cx_cl_display = XOpenDisplay(display_name))
      tree_top = DefaultScreenOfDisplay(cx_cl_display)->root;
    else
      return NULL;
  }

/* Local if we are open as we should be. */

  else if (cx_display) {
    cx_cl_display = cx_display;
    tree_top = cx_rootwindow;
  }

/* Called before initial calls. */

  else {
    printf("gui_init must be called before cx_getcmd\n");
    return NULL;
  }

/* Clear the array and start at top of tree. */

  cx_cl_count = 0;
  cx_command_lines[0] = NULL;

  cx_cl_getcmd(tree_top, 0);

/* Close a remote connection. */

  if (cx_cl_display != cx_display)
    XCloseDisplay(cx_cl_display);

/* We should make an allocated copy of the array for caller,
   so they could release it as they expect to,
   after they release each line. */

  return cx_command_lines;
}
