/* See uidef comments on Active function call,
   and only doing update for changed values.
   These apply here as well.
   Namely we are not done yet.
   
   Also notes for here only:
   Storing a new value lower on the dictionary stack,
   should invalidate (or better just remove) any higher copy,
   since this is meant to update an ancestral value.

   Easy for a single dictionary stack,
   but tricky if we ever become a tree.
   Could let it wait until someone attempts to get the value,
   have them always run down their own version of the stack,
   to make sure that the top is really the good one.
   That gets expensive! */

#include <varargs.h>
#include <string.h>
#include <stdlib.h>
#include <memory.h>
#include <malloc.h>

#include "gui_comm.h"

int uistore(va_alist) va_dcl

/* Update entry in designated level dictionary.

   Entry is a descriptor with name and value or pointer.

   This uses the variable argument list macro package.

   The first argument is the name of the node
   on the active stack to update.

   The second argument is the name of the entry.

   The third is a code for kind of data,
   which determines meaning of last,
   some kind of value. */

{ static char   kws[] = "@(#)uistore.c	1.1 ARGOS 93/12/13 17:37:18";

    long     ltemp;
    double    dtemp;

    char       *svp;

    char       *node;
    char       *ename;
    int       kind;

    int       level;
    UIDATA     *ii;

    va_list     ap;

    va_start(ap);

    node = va_arg(ap, char *);

    ename = va_arg(ap, char *);

    kind = va_arg(ap, int);

/* Search for the named node on the active stack.
   If not found assume late arrival from closing window,
   these should simply be ignored. */

    for (level = Actives;
         level;
         level--)
        if (strcasecmp(Node_Name(Active_Nodes[level]), node) == 0)
            break;

    if (!level) {
#ifdef DEBUG
        if (DEBUG)
            fprintf(DEBUG,
                    "uistore: node %s not found\n", node);
#endif
        return 0;
    }

/* Search for the name only in the designated node dictionary,
   not the entire active dictionary stack. */

    for (ii = Data[level];
       ii;
       ii = Data_Sibling(ii)) {
        if (strcasecmp(ename, Data_Name(ii)) == 0)
            break;
    }

/* Not found we create one and insert to dictionary at this level. */

    if (!ii) {
        ii = (UIDATA *) malloc(sizeof(UIDATA));
	memset(ii, 0x00, sizeof(UIDATA));
        Data_Sibling(ii) = Data[level];
        Data[level] = ii;

        Data_Name(ii) = strdup(ename);
        Data_Kind(ii) = kind;
        Data_SValue(ii) = 0;
    }

/* Sequence number all data creation and updates. */

    Data_SeqStamp(ii) = ++SeqStamp;

    switch (kind) {

    case LONGVAL:
        ltemp = va_arg(ap, long);
	if (Data_Active(ii) && Data_Active(ii)(ii, ltemp) == 0)
	    return;

        Data_LValue(ii) = ltemp;
#ifdef DEBUG
        if (DEBUG)
            fprintf(DEBUG,
            "store LONG %ld %s,%s as %d\n",
            Data_SeqStamp(ii), node, ename, Data_LValue(ii));
#endif
        break;

/* Passed to us as double not float in spite of how we declare it,
   (tried float variable and (float) cast on variable or literal)
   so take it here as a double and then stuff into float. */

    case FLOATVAL:
        dtemp = va_arg(ap, float);
	if (Data_Active(ii) && Data_Active(ii)(ii, dtemp) == 0)
	    return;

        Data_FValue(ii) = dtemp;
#ifdef DEBUG
        if (DEBUG)
            fprintf(DEBUG,
            "store FLOAT %ld %s,%s as %f\n",
            Data_SeqStamp(ii), node, ename, Data_FValue(ii));
#endif
        break;

    case DOUBLEVAL:
        dtemp = va_arg(ap, double);
	if (Data_Active(ii) && Data_Active(ii)(ii, dtemp) == 0)
	    return;

        Data_DValue(ii) = dtemp;
#ifdef DEBUG
        if (DEBUG)
            fprintf(DEBUG,
            "store DOUBLE %ld %s,%s as %lf\n",
            Data_SeqStamp(ii), node, ename, Data_DValue(ii));
#endif
        break;

    case STRINGVAL:
        svp = va_arg(ap, char *);

	if (Data_Active(ii) && Data_Active(ii)(ii, svp) == 0)
	    return;
        
        if (svp != Data_SValue(ii)) {
            if (Data_SValue(ii))
                free(Data_SValue(ii));
            Data_SValue(ii) = strdup(svp);
        }
#ifdef DEBUG
        if (DEBUG)
            fprintf(DEBUG,
            "store STRING %ld %s,%s as %s\n",
            Data_SeqStamp(ii), node, ename, Data_SValue(ii));
#endif
        break;

/* just store a malloc pointer provided by the caller.
   release any previous allocation. */

    case MALLOCVAL:
        svp = va_arg(ap, char *);

	if (Data_Active(ii) && Data_Active(ii)(ii, svp) == 0)
	    return;
        
        if (svp != Data_SValue(ii)) {
            if (Data_SValue(ii))
                free(Data_SValue(ii));
            Data_SValue(ii) = svp;
        }
#ifdef DEBUG
        if (DEBUG)
            fprintf(DEBUG,
            "store MALLOC %ld %s,%s %ld\n",
            Data_SeqStamp(ii), node, ename, Data_SValue(ii));
#endif
        break;

/* pass the size of other (structured) data. */

    case STRUCTVAL:
        ltemp = va_arg(ap, long);
        svp = (va_arg(ap, char *));

	if (Data_Active(ii) && Data_Active(ii)(ii, svp) == 0)
	    return;
        
        if (svp != Data_SValue(ii)) {
            if (Data_SValue(ii))
                free(Data_SValue(ii));
            Data_SValue(ii) = memcpy((void *)malloc(ltemp),
                                     (void *)svp,
                                     (size_t)ltemp);
        }
#ifdef DEBUG
        if (DEBUG)
            fprintf(DEBUG,
            "store STRUCT %ld %s,%s %ld\n",
            Data_SeqStamp(ii), node, ename, Data_SValue(ii));
#endif
        break;

    }

    va_end(ap);

#ifdef DEBUG
    if (DEBUG) fflush(DEBUG);
#endif

    return level;
}
